
from typing import Final

from logging import Logger
from logging import getLogger

from math import pi
from math import sin
from math import atan
from math import cos
from typing import Tuple

# noinspection PyPackageRequirements
from fpdf import FPDF

from pyumldiagrams.IDiagramLine import IDiagramLine
from pyumldiagrams.UnsupportedException import UnsupportedException

from pyumldiagrams.Internal import ArrowPoints
from pyumldiagrams.Internal import DiamondPoints
from pyumldiagrams.Internal import PolygonPoints
from pyumldiagrams.Internal import InternalPosition
from pyumldiagrams.Internal import ScanPoints

from pyumldiagrams.Definitions import LinePositions
from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import Position
from pyumldiagrams.pdf.PdfCommon import Coordinates

from pyumldiagrams.pdf.PdfCommon import PdfCommon


class PdfLine(IDiagramLine):
    """
    This class takes responsibility for drawing the various types of lines within the
    described UML classes.  End users generally do not directly use this class.
    It is split off as part of the separation of responsibility principle.
    """
    INHERITANCE_ARROW_HEIGHT: Final = 10
    DIAMOND_HEIGHT:           Final = 8

    def __init__(self, pdf: FPDF, diagramPadding: DiagramPadding, dpi: int):

        super().__init__(docMaker=pdf, diagramPadding=diagramPadding, dpi=dpi)
        self.logger: Logger = getLogger(__name__)

        # self._pdf: FPDF = pdf
        # self._dpi: int  = dpi
        # self._diagramPadding: diagramPadding  = diagramPadding

    def draw(self, lineDefinition: UmlLineDefinition):
        """
        Draw the line described by the input parameter
        Args:
            lineDefinition:  Describes the line to draw
        """

        linePositions: LinePositions = lineDefinition.linePositions
        lineType:      LineType      = lineDefinition.lineType

        if lineType == LineType.Inheritance:
            self._drawInheritanceArrow(linePositions=linePositions)
        elif lineType == LineType.Composition:
            self._drawCompositionSolidDiamond(linePositions=linePositions)
        elif lineType == LineType.Aggregation:
            self._drawAggregationDiamond(linePositions=linePositions)
        elif lineType == LineType.Association:
            self._drawAssociation(linePositions=linePositions)
        else:
            raise UnsupportedException(f'Line definition type not supported: `{lineType}`')

    def _drawInheritanceArrow(self, linePositions: LinePositions):
        """
        Must account for the margins and gaps between drawn shapes
        Must convert to from screen coordinates to point coordinates
        Draw the arrow first
        Compute the mid-point of the bottom line of the arrow
        That is where the line ends

        Args:
            linePositions - The points that describe the line
        """
        lastIdx:       int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1
        endPoints: Tuple[InternalPosition, InternalPosition] = self.__convertEndPoints(linePositions[beforeLastIdx], linePositions[lastIdx])

        convertedSrc:  InternalPosition = endPoints[0]
        convertedDst:  InternalPosition = endPoints[1]

        points: ArrowPoints = self.__computeTheArrowVertices(convertedSrc, convertedDst)
        self.__drawPolygon(points)

        newEndPoint: InternalPosition = self.__computeMidPointOfBottomLine(points[0], points[2])

        self.__finishDrawingLine(linePositions=linePositions, newEndPoint=newEndPoint)

    def _drawCompositionSolidDiamond(self, linePositions: LinePositions):

        lastIdx:       int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1
        endPoints: Tuple[InternalPosition, InternalPosition] = self.__convertEndPoints(linePositions[beforeLastIdx], linePositions[lastIdx])

        convertedSrc: InternalPosition = endPoints[0]
        convertedDst: InternalPosition = endPoints[1]

        points: DiamondPoints = self.__computeDiamondVertices(convertedSrc, convertedDst)
        self.__drawPolygon(points)
        self.__fillInDiamond(points)

        newEndPoint: InternalPosition = points[3]

        self.__finishDrawingLine(linePositions=linePositions, newEndPoint=newEndPoint)

    def _drawAggregationDiamond(self, linePositions: LinePositions):

        lastIdx:   int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1

        endPoints: Tuple[InternalPosition, InternalPosition] = self.__convertEndPoints(linePositions[beforeLastIdx], linePositions[lastIdx])

        convertedSrc: InternalPosition = endPoints[0]
        convertedDst: InternalPosition = endPoints[1]

        points: ArrowPoints = self.__computeDiamondVertices(convertedSrc, convertedDst)
        self.__drawPolygon(points)

        newEndPoint: InternalPosition = points[3]

        self.__finishDrawingLine(linePositions=linePositions, newEndPoint=newEndPoint)

    def _drawAssociation(self, linePositions: LinePositions):

        verticalGap:   int  = self._diagramPadding.verticalGap
        horizontalGap: int  = self._diagramPadding.horizontalGap
        dpi:           int  = self._dpi
        docMaker:      FPDF = self._docMaker

        numPositions: int      = len(linePositions)
        for idx in range(numPositions):

            nextIdx:    int      = idx + 1
            currentPos: Position = linePositions[idx]
            if nextIdx == numPositions:
                break
            nextPos: Position = linePositions[nextIdx]

            currentCoordinates: Coordinates = PdfCommon.convertPosition(pos=currentPos, dpi=dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)
            nextCoordinates:    Coordinates = PdfCommon.convertPosition(pos=nextPos, dpi=dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)

            docMaker.line(x1=currentCoordinates.x, y1=currentCoordinates.y, x2=nextCoordinates.x, y2=nextCoordinates.y)

    def __convertEndPoints(self, src: Position, dst: Position) -> Tuple[InternalPosition, InternalPosition]:

        verticalGap:   int = self._diagramPadding.verticalGap
        horizontalGap: int = self._diagramPadding.horizontalGap

        sourceCoordinates:      Coordinates = PdfCommon.convertPosition(pos=src, dpi=self._dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)
        destinationCoordinates: Coordinates = PdfCommon.convertPosition(pos=dst, dpi=self._dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)

        convertedSrc: InternalPosition = InternalPosition(sourceCoordinates.x, sourceCoordinates.y)
        convertedDst: InternalPosition = InternalPosition(destinationCoordinates.x, destinationCoordinates.y)

        return convertedSrc, convertedDst

    def __computeTheArrowVertices(self, src: InternalPosition, dst: InternalPosition)  -> ArrowPoints:
        """
        Draw an arrow at the end of the line source-destination.

        Args:
            src:  points of the segment
            dst:  points of the segment

        Returns:
            A list of positions that describes a diamond to draw
        """
        # x1: float = src.x
        # y1: float = src.y
        # x2: float = dst.x
        # y2: float = dst.y
        #
        # deltaX: float = x2 - x1
        # deltaY: float = y2 - y1
        deltaX, deltaY = self.__computeDeltaXDeltaY(src, dst)
        if abs(deltaX) < 0.01:   # vertical segment
            if deltaY > 0:
                alpha = -pi/2
            else:
                alpha = pi/2
        else:
            if deltaX == 0:
                alpha = pi/2
            else:
                alpha = atan(deltaY/deltaX)
        if deltaX > 0:
            alpha += pi

        pi_6: float = pi/6      # radians for 30 degree angle

        alpha1: float = alpha + pi_6
        alpha2: float = alpha - pi_6
        size:   float = PdfLine.INHERITANCE_ARROW_HEIGHT
        x2: int = dst.x
        y2: int = dst.y
        #
        # The names for the left and right points are correct for upward facing arrows
        # They are inverted for downward facing arrows
        #
        arrowTip:   InternalPosition = InternalPosition(x2, y2)
        rightPoint: InternalPosition = InternalPosition(x2 + round(size * cos(alpha1)), y2 + round(size * sin(alpha1)))
        leftPoint:  InternalPosition = InternalPosition(x2 + round(size * cos(alpha2)), y2 + round(size * sin(alpha2)))

        points: ArrowPoints = [rightPoint, arrowTip, leftPoint]

        return points

    def __computeDiamondVertices(self, src: InternalPosition, dst: InternalPosition) -> DiamondPoints:
        """
        Args:
            src:
            dst:
        """
        pi_6: float = pi/6     # radians for 30 degree angle
        x2:   int = dst.x
        y2:   int = dst.y

        deltaX, deltaY = self.__computeDeltaXDeltaY(src, dst)

        if abs(deltaX) < 0.01:  # vertical segment
            if deltaY > 0:
                alpha = -pi/2
            else:
                alpha = pi/2
        else:
            if deltaX == 0:
                if deltaY > 0:
                    alpha = pi/2
                else:
                    alpha = 3 * pi / 2
            else:
                alpha = atan(deltaY/deltaX)
        if deltaX > 0:
            alpha += pi

        alpha1: float = alpha + pi_6
        alpha2: float = alpha - pi_6
        size:   int   = PdfLine.DIAMOND_HEIGHT

        # noinspection PyListCreation
        points: DiamondPoints = []

        points.append((InternalPosition(x2 + round(size * cos(alpha1)), y2 + round(size * sin(alpha1)))))
        points.append(InternalPosition(x2, y2))
        points.append(InternalPosition(x2 + round(size * cos(alpha2)), y2 + round(size * sin(alpha2))))
        points.append(InternalPosition(x2 + 2 * round(size * cos(alpha)), y2 + 2 * round(size * sin(alpha))))

        return points

    def __drawPolygon(self, points: PolygonPoints):

        pdf: FPDF = self._docMaker
        ptNumber: int = 0
        for point in points:

            x1: int = point.x
            y1: int = point.y

            if ptNumber == len(points) - 1:
                nextPoint = points[0]
                x2: int = nextPoint.x
                y2: int = nextPoint.y
                pdf.line(x1, y1, x2, y2)
                break
            else:
                nextPoint = points[ptNumber + 1]
                x2 = nextPoint.x
                y2 = nextPoint.y
                pdf.line(x1, y1, x2, y2)

            ptNumber += 1

    def __computeMidPointOfBottomLine(self, startPos: InternalPosition, endPos: InternalPosition) -> InternalPosition:
        """
        These two coordinates are the two end-points of the bottom leg of the inheritance arrow
        midPoint = (x1+x2/2, y1+y2/2)

        Args:
            startPos: start of line
            endPos:   end of line

        Returns:  Midpoint between startPos - endPos

        """
        x1: int = startPos.x
        y1: int = startPos.y
        x2: int = endPos.x
        y2: int = endPos.y

        midX: int = (x1 + x2) // 2
        midY: int = (y1 + y2) // 2

        return InternalPosition(midX, midY)

    def __computeDeltaXDeltaY(self, src: InternalPosition, dst: InternalPosition) -> Tuple[float, float]:

        x1: float = src.x
        y1: float = src.y
        x2: float = dst.x
        y2: float = dst.y

        deltaX: float = x2 - x1
        deltaY: float = y2 - y1

        return deltaX, deltaY

    def __fillInDiamond(self, points: DiamondPoints):
        """

        Args:
            points:  The polygon that defines the composition diamond

        """
        scanPoints: ScanPoints = PdfCommon.buildScanPoints(points)

        startX: int = scanPoints.startScan.x
        startY: int = scanPoints.startScan.y

        endX: int = scanPoints.endScan.x
        endY: int = scanPoints.endScan.y

        x = startX
        while x <= endX:
            y = startY
            while y <= endY:
                if PdfCommon.pointInsidePolygon(pos=InternalPosition(x, y), polygon=points):
                    self._docMaker.line(x1=x, y1=y, x2=x, y2=y)
                y += 1
            x += 1

    def __finishDrawingLine(self, linePositions: LinePositions, newEndPoint: InternalPosition):

        linePositionsCopy: LinePositions = linePositions[:-1]  # Makes a copy; remove last one

        verticalGap:   int  = self._diagramPadding.verticalGap
        horizontalGap: int  = self._diagramPadding.horizontalGap
        dpi:           int  = self._dpi
        docMaker:      FPDF = self._docMaker
        #
        # Ok, ok, I get it.  This is not a Pythonic 'for' loop.  But, I am not a purist
        #
        numPositions: int      = len(linePositionsCopy)
        currentPos:   Position = linePositionsCopy[0]
        for idx in range(numPositions):

            nextIdx: int = idx + 1
            currentPos = linePositionsCopy[idx]
            if nextIdx == numPositions:
                break
            nextPos: Position = linePositionsCopy[nextIdx]

            currentCoordinates: Coordinates = PdfCommon.convertPosition(pos=currentPos, dpi=dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)
            nextCoordinates:    Coordinates = PdfCommon.convertPosition(pos=nextPos, dpi=dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)

            docMaker.line(x1=currentCoordinates.x, y1=currentCoordinates.y, x2=nextCoordinates.x, y2=nextCoordinates.y)

        currentCoordinates = PdfCommon.convertPosition(pos=currentPos, dpi=dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)

        docMaker.line(x1=currentCoordinates.x, y1=currentCoordinates.y, x2=newEndPoint.x, y2=newEndPoint.y)




from typing import Final

from logging import Logger
from logging import getLogger

from fpdf import FPDF


class FPDFExtended(FPDF):

    DEFAULT_PAGE_WIDTH:  Final = 3000     # points
    DEFAULT_PAGE_HEIGHT: Final = 1500     # points

    def __init__(self, headerText: str = ''):

        super().__init__(orientation='L', unit='pt', format=(FPDFExtended.DEFAULT_PAGE_HEIGHT, FPDFExtended.DEFAULT_PAGE_WIDTH))

        self.logger: Logger = getLogger(__name__)

        self._headerText: str = headerText

    def header(self):

        self.set_font('Arial', 'B', 15)
        # Move to the right
        self.cell(80)

        textWidth: int = self.get_string_width(self._headerText)
        # self.cell(w=textWidth, h=10, txt=f'{self._headerText}', border=0, ln=0, align='L')
        self.cell(w=textWidth, h=10, txt=f'{self._headerText}', border=0, align='L')
        # Line break
        self.ln(20)

    def footer(self):

        self.set_y(-15)

        self.set_font('Arial', 'I', 15)
        # Print Left Aligned page number
        # self.cell(0, 10, 'Page %s' % self.page_no(), 0, 0, 'L')
        pageNum: str = f'Page {self.page_no()}'
        self.cell(w=0, h=10, txt=pageNum, border=0, align='L')




from typing import List
from typing import NewType

from enum import Enum

from dataclasses import dataclass
from dataclasses import field

from pyumldiagrams.Defaults import TOP_MARGIN
from pyumldiagrams.Defaults import LEFT_MARGIN
from pyumldiagrams.Defaults import DEFAULT_HORIZONTAL_GAP
from pyumldiagrams.Defaults import DEFAULT_VERTICAL_GAP
from pyumldiagrams.UnsupportedException import UnsupportedException


ClassName = NewType('ClassName', str)


@dataclass
class Position:
    """
    The x and y coordinates are in screen/display resolution.  Diagramming modules may
    convert these to appropriate positions based on the rendering technology.
    """
    x: int = 0
    """
    The x-axis (horizontal) abscissa
    """
    y: int = 0
    """
    The y-axis (vertical) ordinate
    """


def createPositionFactory() -> Position:
    return Position()


@dataclass
class DiagramPadding:
    """
    todo::  These should move to the Internal package
    """

    topMargin:  int = TOP_MARGIN
    """
    The diagram's observed top margin.  See `pyumldiagrams.Defaults.TOP_MARGIN`
    """
    leftMargin: int = LEFT_MARGIN
    """
    The diagram's observed left margin.  See `pyumldiagrams.Defaults.LEFT_MARGIN`
    """

    horizontalGap: int = DEFAULT_HORIZONTAL_GAP
    """
    The horizontal gap between UML graphics added to the layout in addition to the gap imposed 
    by the actual graphics positions. See `pyumldiagrams.Defaults.DEFAULT_HORIZONTAL_GAP`
    """
    verticalGap:   int = DEFAULT_VERTICAL_GAP
    """
    The vertical gap between UML graphics added to the layout in addition to the gap imposed 
    by the actual graphics positions.  See `pyumldiagrams.Defaults.DEFAULT_VERTICAL_GAP`
    """


@dataclass
class Size:
    """
    Defines the size of the input UML definitions;
    """
    width:  int = 100
    """
    The width of a shape
    """
    height: int = 100
    """
    The height of the shape
    """


def createSizeFactory() -> Size:
    return Size()


class DefinitionType(Enum):
    """
    Defines the visibility of either methods or fields
    """
    Public    = '+'
    Private   = '-'
    Protected = '#'


@dataclass
class BaseDefinition:

    __slots__ = ['name']
    name: str
    """
    The name associated with the definition.
    """


@dataclass
class ParameterDefinition(BaseDefinition):
    """
    Defines a single parameter for a method
    """
    parameterType: str = ''
    """
    A string that describes the parameter type
    """
    defaultValue:  str = ''
    """
    A string that describes a parameter default value
    """


Parameters = NewType('Parameters', List[ParameterDefinition])
"""
Syntactic sugar to define a list of parameters.  
"""


class DisplayMethodParameters(Enum):

    DISPLAY        = 'Display'
    DO_NOT_DISPLAY = 'DoNotDisplay'
    UNSPECIFIED    = 'Unspecified'


def createParametersFactory() -> Parameters:
    return Parameters([])

@dataclass
class MethodDefinition(BaseDefinition):
    """
    Defines a single method in a UML class
    """
    visibility: DefinitionType = DefinitionType.Public
    """
    Defines the method visibility.  See `DefinitionType`
    """
    returnType: str = ''
    """
    Defines the method return type.
    """
    parameters: Parameters = field(default_factory=createParametersFactory)
    """
    Define the parameters for a particular method
    """


Methods = NewType('Methods', List[MethodDefinition])
"""
Syntactic sugar to define a list of methods.
"""


def createMethodsFactory() -> Methods:
    return Methods([])


@dataclass
class FieldDefinition(ParameterDefinition):
    """
    Defines a single instance variable;  Seems funny to inherit from a
    parameter definition.
    """
    visibility: DefinitionType = DefinitionType.Public
    """
    Defines the field visibility.  See `DefinitionType`
    """


Fields = NewType('Fields', List[FieldDefinition])


def createFieldsFactory() -> Fields:
    return Fields([])


@dataclass
class ClassDefinition(BaseDefinition):
    """ The class definition.  Currently, does not support instance properties.
    """
    size:     Size     = field(default_factory=createSizeFactory)
    """
    The size of UML class symbol.  See `Size`
    """
    position: Position = field(default_factory=createPositionFactory)
    """
    The position of the UML class symbol.  See `Position`
    """
    methods: Methods   = field(default_factory=createMethodsFactory)
    """
    The list of methods this class implements.  
    """
    fields:  Fields    = field(default_factory=createFieldsFactory)
    """
    The list of instance variables this class defines.
    """
    displayStereotype:  bool = True
    """
    If true display the class stereotype when drawing the class diagram
    """
    displayMethods:     bool = True
    """
    If True display the class methods
    """
    displayFields:      bool = True
    """
    If True display the class instance variables
    """
    displayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.UNSPECIFIED
    """
    If True display the method parameters;  If UNSPECIFIED defer to global
    """


ClassDefinitions = List[ClassDefinition]


class LineType(Enum):
    """
    The type of UML line you wish to draw.  Currently, bare associations are not supported.
    """
    Inheritance     = 0
    Aggregation     = 1
    Composition     = 3
    Association     = 7
    NoteAssociation = 9

    @staticmethod
    def toEnum(strValue: str) -> 'LineType':
        """
        Converts the input string to the line type enum
        Args:
            strValue:   The serialized string representation

        Returns:  The line type enumeration
        """
        canonicalStr: str = strValue.lower().strip(' ')
        if canonicalStr == 'aggregation':
            return LineType.Aggregation
        elif canonicalStr == 'composition':
            return LineType.Composition
        elif canonicalStr == 'inheritance':
            return LineType.Inheritance
        elif canonicalStr == 'association':
            return LineType.Association
        elif canonicalStr == 'noteAssociation':
            return LineType.NoteAssociation
        else:
            raise UnsupportedException(f'Do not handle LineType {canonicalStr}')


LinePositions = List[Position]


@dataclass
class LineDefinition:
    """
    Defines a line between many points;  Index 0 the start of the line;  That last point
    is the end of the line
    """
    linePositions: LinePositions


@dataclass
class UmlLineDefinition(LineDefinition):
    """
    A UML Line definition includes its type
    """
    lineType: LineType
    """
    The UML line type  See `LineType`.
    """


UmlLineDefinitions = List[UmlLineDefinition]
"""
Syntactic sugar to define a list of UML Lines.
"""


class RenderStyle(Enum):
    """
    An enumeration that determines how to draw various UML and other graphical elements
    """
    Draw     = 'D'
    """
    Just draw the outline
    """
    Fill     = 'F'
    """
    Just fill in the area associated with the shape
    """
    DrawFill = 'DF'
    """
    Do both when drawing the UML shape or figure
    """


@dataclass
class RectangleDefinition:
    """
    Defines a rectangle
    """

    renderStyle: RenderStyle = RenderStyle.Draw
    """
    How to draw the rectangle.  See `RenderStyle`
    """
    position:    Position    = field(default_factory=createPositionFactory)
    """
    Where to put the rectangle.  See `Position`
    """
    size:        Size        = field(default_factory=createSizeFactory)
    """
    The rectangle size.  See `Size`
    """


@dataclass
class EllipseDefinition(RectangleDefinition):
    """
    This is just typing syntactical sugar on how to define an Ellipse.
    """
    pass




from dataclasses import dataclass
from dataclasses import field

from pyumldiagrams.Common import Common
from pyumldiagrams.Definitions import Position

from pyumldiagrams.Defaults import LEFT_MARGIN
from pyumldiagrams.Defaults import TOP_MARGIN

from pyumldiagrams.Internal import InternalPosition
from pyumldiagrams.Internal import PolygonPoints
from pyumldiagrams.Internal import ScanPoints


@dataclass
class Coordinates:
    x: int = 0
    y: int = 0


def createCoordinatesFactory() -> Coordinates:
    return Coordinates()


@dataclass
class Dimensions:
    width:  int = 0
    height: int = 0


def createDimensionsFactory() -> Dimensions:
    return Dimensions()


@dataclass
class PdfShapeDefinition:
    coordinates: Coordinates = field(default_factory=createCoordinatesFactory)
    dimensions:  Dimensions  = field(default_factory=createDimensionsFactory)


class PdfCommon(Common):

    @classmethod
    def toPdfPoints(cls, pixelNumber: float, dpi: int) -> int:
        """

        points = pixels * 72 / DPI

        Args:
            pixelNumber:  From the display
            dpi:  dots per inch of source display

        Returns:  A pdf point value to use to position on a generated document

        """
        points: int = int((pixelNumber * 72)) // dpi

        return points

    @classmethod
    def convertPosition(cls, pos: Position, dpi: int, verticalGap: int, horizontalGap: int) -> Coordinates:

        x: int = PdfCommon.toPdfPoints(pos.x, dpi) + LEFT_MARGIN + verticalGap
        y: int = PdfCommon.toPdfPoints(pos.y, dpi) + TOP_MARGIN + horizontalGap

        return Coordinates(x=x, y=y)

    @classmethod
    def pointInsidePolygon(cls, pos: InternalPosition, polygon: PolygonPoints) -> bool:
        """
        Based on this: http://www.ariel.com.au/a/python-point-int-poly.html

        Args:
            pos: The position to check
            polygon: The polygon

        Returns: True if it is, else False
        """
        x: int = pos.x
        y: int = pos.y
        n: int   = len(polygon)

        inside: bool = False

        p1: InternalPosition = polygon[0]
        p1x: float = p1.x
        p1y: float = p1.y

        for i in range(n + 1):
            p2: InternalPosition = polygon[i % n]
            p2x: float = p2.x
            p2y: float = p2.y

            if y > min(p1y, p2y):
                if y <= max(p1y, p2y):
                    if x <= max(p1x, p2x):
                        xIntersection: float = 0
                        if p1y != p2y:
                            xIntersection = (y - p1y) * (p2x - p1x) // (p2y - p1y) + p1x
                        if p1x == p2x or x <= xIntersection:
                            inside = not inside

            p1x = p2x
            p1y = p2y

        return inside

    @classmethod
    def buildScanPoints(cls, points: PolygonPoints) -> ScanPoints:

        minX: int = points[0].x
        maxX: int = points[0].y
        minY: int = points[0].x
        maxY: int = points[0].y

        for point in points:
            currX: int = point.x
            currY: int = point.y
            if currX < minX:
                minX = currX
            if currX > maxX:
                maxX = currX

            if currY < minY:
                minY = currY
            if currY > maxY:
                maxY = currY

        scanPoints: ScanPoints = ScanPoints()

        scanPoints.startScan = InternalPosition(minX, minY)
        scanPoints.endScan   = InternalPosition(maxX, maxY)

        return scanPoints




from typing import List
from typing import cast
from typing import Final

from logging import Logger
from logging import getLogger

from datetime import datetime

from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import DisplayMethodParameters
from pyumldiagrams.Definitions import EllipseDefinition
from pyumldiagrams.Definitions import FieldDefinition
from pyumldiagrams.Definitions import MethodDefinition
from pyumldiagrams.Definitions import ParameterDefinition
from pyumldiagrams.Definitions import RectangleDefinition
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import Fields
from pyumldiagrams.Definitions import Methods


class BaseDiagram:
    """
    Always lays out in portrait mode.  Currently only supports UML classes with methods.  Only supports
    inheritance, composition, and aggregation lines.

    You are allowed to set the gap between UML classes both horizontally and vertically.  Also, you are allowed to
    specify the text font size
    """

    MethodsRepr = List[str]
    FieldsRepr  = List[str]

    DEFAULT_FONT_SIZE: Final = 10
    HEADER_FONT_SIZE:  Final = 14
    # noinspection SpellCheckingInspection
    RESOURCE_ENV_VAR:  Final = 'RESOURCEPATH'

    clsLogger: Logger = getLogger(__name__)

    def __init__(self, fileName: str,
                 docDisplayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.DISPLAY,
                 dpi: int = 0,
                 headerText: str = ''
                 ):
        """

        Args:
            fileName:   Fully qualified file name

            docDisplayMethodParameters: global flag to determine whether or display a method's parameters

            dpi: dots per inch for the display we are mapping from;
            Some diagramming documents may not need a value for this since they map directly to display device

            headerText:  The header to place on the page
        """

        self._fileName:   str = fileName

        self._docDisplayMethodParameters: DisplayMethodParameters = docDisplayMethodParameters

        self._dpi:        int = dpi
        self._headerText: str = headerText
        self._fontSize:   int = BaseDiagram.DEFAULT_FONT_SIZE

        self._softwareNameVersion: str = ''
        self._diagramPadding:      DiagramPadding = DiagramPadding()

    @property
    def fontSize(self) -> int:
        """
        The font size to use in the generated UML diagram.  If unchanged the value is `pyumldiagrams.BaseDiagram.BaseDiagram.DEFAULT_FONT_SIZE`
        """
        return self._fontSize

    @fontSize.setter
    def fontSize(self, newSize: int):
        self._fontSize = newSize

    @property
    def horizontalGap(self) -> int:
        """
        The horizontal gap between UML elements.  If not set then the value is
        `pyumldiagrams.Defaults.DEFAULT_HORIZONTAL_GAP`
        """
        return self._diagramPadding.horizontalGap

    @horizontalGap.setter
    def horizontalGap(self, newValue: int):
        self._diagramPadding.horizontalGap = newValue

    @property
    def verticalGap(self) -> int:
        """
        The vertical gap between UML elements.  If not set then the value is
        `pyumldiagrams.Defaults.DEFAULT_VERTICAL_GAP`
        """
        return self._diagramPadding.verticalGap

    @verticalGap.setter
    def verticalGap(self, newValue):
        self._diagramPadding.verticalGap = newValue

    @property
    def headerText(self) -> str:
        """
        The text to display as the header on the generated document
        """
        return self._headerText

    @headerText.setter
    def headerText(self, newValue: str):
        self._headerText = newValue

    @property
    def softwareNameVersion(self) -> str:
        """
        Used to place the software name and version of the application using this library onto the
        generated image EXIF metadata
        """
        return self._softwareNameVersion

    @softwareNameVersion.setter
    def softwareNameVersion(self, newValue: str):
        self._softwareNameVersion = newValue

    @property
    def docTimeStamp(self) -> datetime:
        """
        Must be overridden by implementors
        """
        return cast(datetime, None)

    @docTimeStamp.setter
    def docTimeStamp(self, timeStamp: datetime):
        """
        Must be overridden by implementors

        """
        pass

    def retrieveResourcePath(self, bareFileName: str) -> str:
        """
        Must be overridden by implementors

        Args:
            bareFileName:

        Returns: a fully qualified name
        """
        return ''

    def drawClass(self, classDefinition: ClassDefinition):
        """
        Draw the class diagram defined by the input
        Must be overridden by implementors

        Args:
            classDefinition:    The class definition
        """
        pass

    def drawUmlLine(self, lineDefinition: UmlLineDefinition):
        """
        Draw the inheritance, aggregation, or composition lines that describe the relationships
        between the UML classes

        Must be overridden by implementors

        Args:
            lineDefinition:   A UML Line definition
        """
        pass

    def drawEllipse(self, definition: EllipseDefinition):
        """
        Draw a general purpose ellipse

        Args:
            definition:     It's definition
        """
        pass

    def drawRectangle(self, definition: RectangleDefinition):
        """
        Draw a general purpose rectangle

        Args:
            definition:  The rectangle definition
        """
        pass

    def write(self):
        """
        Call this method when you are done with placing the diagram onto a document.
        Must be overridden by implementors
        """
        pass

    def _buildMethods(self, methods: Methods, displayParameters: DisplayMethodParameters) -> MethodsRepr:
        """

        Args:
            methods: The method definitions
            displayParameters:  Determines how to build the method parameter signature

        Returns:  The text version of each method as a list of method representations
        """

        methodReprs: BaseDiagram.MethodsRepr = []

        for methodDef in methods:

            methodRepr: str = self._buildMethod(methodDef, displayParameters)
            methodReprs.append(methodRepr)

        return methodReprs

    def _buildMethod(self, methodDef: MethodDefinition, displayParameters: DisplayMethodParameters) -> str:

        if methodDef.visibility is None:
            methodRepr: str = f'{methodDef.name}'
        else:
            methodRepr = f'{methodDef.visibility.value} {methodDef.name}'

        paramRepr: str = ''

        if displayParameters == DisplayMethodParameters.DISPLAY:

            self.clsLogger.debug(f'{methodDef.name} - {len(methodDef.parameters)=}')

            paramRepr = self.__generateParametersString(methodDef, paramRepr)
        elif displayParameters == DisplayMethodParameters.UNSPECIFIED:
            if self._docDisplayMethodParameters == DisplayMethodParameters.DISPLAY:
                paramRepr = self.__generateParametersString(methodDef, paramRepr)

        methodRepr = f'{methodRepr}({paramRepr})'

        return methodRepr

    def _buildFields(self, fields: Fields) -> FieldsRepr:

        fieldsRepr: BaseDiagram.FieldsRepr = []

        for fieldDef in fields:
            fieldRepr: str = self._buildField(fieldDef)
            fieldsRepr.append(fieldRepr)

        return fieldsRepr

    def _buildField(self, fieldDef: FieldDefinition) -> str:

        fieldRepr: str = f'{fieldDef.name}'

        if fieldDef.parameterType != '' and fieldDef.parameterType is not None:
            fieldRepr = f'{fieldRepr}: {fieldDef.parameterType}'

        if fieldDef.defaultValue != '' and fieldDef.defaultValue is not None:
            fieldRepr = f'{fieldRepr} = {fieldDef.defaultValue}'

        return fieldRepr

    def __generateParametersString(self, methodDef, paramRepr):

        nParams:  int = len(methodDef.parameters)
        paramNum: int = 0

        for parameterDef in methodDef.parameters:
            parameterDef = cast(ParameterDefinition, parameterDef)
            paramNum += 1

            paramRepr = f'{paramRepr}{parameterDef.name}'

            if parameterDef.parameterType is None or len(parameterDef.parameterType) == 0:
                paramRepr = f'{paramRepr}'
            else:
                paramRepr = f'{paramRepr}: {parameterDef.parameterType}'

            if parameterDef.defaultValue is None or len(parameterDef.defaultValue) == 0:
                paramRepr = f'{paramRepr}'
            else:
                paramRepr = f'{paramRepr}={parameterDef.defaultValue}'

            if paramNum == nParams:
                paramRepr = f'{paramRepr}'
            else:
                paramRepr = f'{paramRepr}, '

        return paramRepr




from pyumldiagrams.Common import Common
from pyumldiagrams.Defaults import LEFT_MARGIN
from pyumldiagrams.Defaults import TOP_MARGIN

from pyumldiagrams.Definitions import Position
from pyumldiagrams.Internal import InternalPosition


class ImageCommon(Common):

    @classmethod
    def toInternal(cls, position: Position, verticalGap: int, horizontalGap: int) -> InternalPosition:
        """
        Assumes a 1 to 1 relationship between display device and the image we are generating.

        Args:
            position:  The original position

            verticalGap:  Account for the vertical gap on the X-axis

            horizontalGap:  Account for the horizontal gap on the Y-axis

        Returns:  The new position adjust for margins and gaps
        """
        adjustedX: int = position.x + LEFT_MARGIN + verticalGap
        adjustedY: int = position.y + TOP_MARGIN  + horizontalGap

        return InternalPosition(x=adjustedX, y=adjustedY)



"""
Sensible defaults for this library
"""
from pyumldiagrams.image.ImageFormat import ImageFormat

LEFT_MARGIN: int = 8
"""
The left margin added to the layout on the UML diagram.  For certain document types
this is in adjusted units
"""
TOP_MARGIN:  int = 8
"""
The top margin added to the layout on the UML diagram.  For certain document types
this is in adjusted units
"""

DEFAULT_HORIZONTAL_GAP: int = 60
"""
The horizontal gap between UML graphics added to the layout in addition to the gap imposed 
by the actual graphics positions
"""
DEFAULT_VERTICAL_GAP:   int = 60
"""
The vertical gap between UML graphics added to the layout in addition to the gap imposed 
by the actual graphics positions
"""

DEFAULT_IMAGE_WIDTH:  int = 1280
"""
The image width
"""
DEFAULT_IMAGE_HEIGHT: int = 1024
"""
The image height"
"""

DEFAULT_LINE_WIDTH: int = 1
"""
The line width
"""

DEFAULT_FILE_NAME: str = 'PyutExport'
"""
The file name
"""

DEFAULT_IMAGE_FORMAT: ImageFormat = ImageFormat.PNG
"""
The Image format

Please don't presume what I am offended by.  Sticks and stones may broke my bones, but words will never hurt me
"""




from typing import Any
from typing import List
from typing import NewType
from typing import Union
from typing import Final

from logging import Logger
from logging import getLogger

from PIL.ImageDraw import ImageDraw

from pyumldiagrams.image.ImageCommon import ImageCommon

from pyumldiagrams.IDiagramLine import IDiagramLine
from pyumldiagrams.UnsupportedException import UnsupportedException

from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import LinePositions

from pyumldiagrams.Internal import ArrowPoints
from pyumldiagrams.Internal import DiamondPoints
from pyumldiagrams.Internal import InternalPosition

PILPoints     = NewType('PILPoints', List[int])
PolygonPoints = NewType('PolygonPoints', List[int])


class ImageLine(IDiagramLine):

    DEFAULT_LINE_COLOR: Final = 'Black'
    LINE_WIDTH:         Final = 1

    def __init__(self, docWriter: Any, diagramPadding: DiagramPadding):

        super().__init__(docMaker=docWriter, diagramPadding=diagramPadding, dpi=0)

        self.logger: Logger = getLogger(__name__)

        self._imgDraw: ImageDraw = docWriter

    def draw(self, lineDefinition: UmlLineDefinition):
        """
        Draw the line described by the input parameter

        Args:
            lineDefinition:  Describes the line to draw
        """
        linePositions: LinePositions = lineDefinition.linePositions
        lineType:      LineType      = lineDefinition.lineType

        if lineType == LineType.Inheritance:
            self._drawInheritanceArrow(linePositions=linePositions)
        elif lineType == LineType.Composition:
            self._drawCompositionSolidDiamond(linePositions=linePositions)
        elif lineType == LineType.Aggregation:
            self._drawAggregationDiamond(linePositions=linePositions)
        elif lineType == LineType.Association:
            self._drawAssociation(linePositions=linePositions)
        else:
            raise UnsupportedException(f'Line definition type not supported: `{lineType}`')

    def _drawInheritanceArrow(self, linePositions: LinePositions):
        """
        Must account for the margins and gaps between drawn shapes
        Must convert from screen coordinates to point coordinates
        Draw the arrow first
        Compute the mid-point of the bottom line of the arrow
        That is where the line ends

        Args:
            linePositions  The points that describe the line
        """
        lastIdx:       int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1
        internalSrc: InternalPosition = self.__toInternal(linePositions[beforeLastIdx])
        internalDst: InternalPosition = self.__toInternal(linePositions[lastIdx])

        points:  ArrowPoints   = ImageCommon.computeTheArrowVertices(internalSrc, internalDst)
        polygon: PolygonPoints = self.__toPolygonPoints(points)

        self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR)

        newEndPoint: InternalPosition = ImageCommon.computeMidPointOfBottomLine(points[0], points[2])
        xy:          PILPoints        = self.__toPILPoints(linePositions=linePositions, newEndPoint=newEndPoint)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

    def _drawCompositionSolidDiamond(self, linePositions: LinePositions):

        lastIdx:       int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1
        internalSrc:  InternalPosition = self.__toInternal(linePositions[beforeLastIdx])
        internalDst:  InternalPosition = self.__toInternal(linePositions[lastIdx])

        points:  DiamondPoints = ImageCommon.computeDiamondVertices(internalSrc, internalDst)
        polygon: PolygonPoints = self.__toPolygonPoints(points)

        self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR)

        newEndPoint: InternalPosition = points[3]
        xy:          PILPoints        = self.__toPILPoints(linePositions=linePositions, newEndPoint=newEndPoint)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

    def _drawAggregationDiamond(self, linePositions: LinePositions):

        lastIdx:       int = len(linePositions) - 1
        beforeLastIdx: int = lastIdx - 1
        internalSrc:  InternalPosition = self.__toInternal(linePositions[beforeLastIdx])
        internalDst:  InternalPosition = self.__toInternal(linePositions[lastIdx])

        points:  DiamondPoints = ImageCommon.computeDiamondVertices(internalSrc, internalDst)
        polygon: PolygonPoints = self.__toPolygonPoints(points)

        self._imgDraw.polygon(xy=polygon, outline=ImageLine.DEFAULT_LINE_COLOR, fill='black')

        newEndPoint: InternalPosition = points[3]
        xy:          PILPoints        = self.__toPILPoints(linePositions=linePositions, newEndPoint=newEndPoint)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

    def _drawAssociation(self, linePositions: LinePositions):

        lastIdx:     int              = len(linePositions) - 1
        internalDst: InternalPosition = self.__toInternal(linePositions[lastIdx])
        xy:          PILPoints        = self.__toPILPoints(linePositions=linePositions, newEndPoint=internalDst)

        self._imgDraw.line(xy=xy, fill=ImageLine.DEFAULT_LINE_COLOR, width=ImageLine.LINE_WIDTH)

    def __toInternal(self, position: Position) -> InternalPosition:

        verticalGap:   int = self._diagramPadding.verticalGap
        horizontalGap: int = self._diagramPadding.horizontalGap

        iPos: InternalPosition = ImageCommon.toInternal(position, verticalGap=verticalGap, horizontalGap=horizontalGap)

        return iPos

    def __toPolygonPoints(self, points: Union[ArrowPoints, DiamondPoints]) -> PolygonPoints:

        polygon: PolygonPoints = PolygonPoints([])

        for point in points:
            polygon.append(int(point.x))
            polygon.append(int(point.y))

        return polygon

    def __toPILPoints(self, linePositions: LinePositions, newEndPoint: InternalPosition) -> PILPoints:

        linePositionsCopy: LinePositions = linePositions[:-1]  # Makes a copy

        xy: PILPoints = PILPoints([])
        for externalPosition in linePositionsCopy:
            internalPosition: InternalPosition = self.__toInternal(externalPosition)
            xy.append(internalPosition.x)
            xy.append(internalPosition.y)

        xy.append(newEndPoint.x)
        xy.append(newEndPoint.y)

        return xy




from typing import List
from typing import Union

from dataclasses import dataclass
from dataclasses import field


@dataclass(eq=True)
class InternalPosition:
    """
    The x and y coordinates are relative to the diagramming method.  For pdf
    documents they are in points.  For images, they are in pixels.  In all cases, the position is
    adjusted for left and right margins plus vertical and horizontal gaps.
    """
    x: int = 0
    y: int = 0


def createInternalPositionFactory() -> InternalPosition:
    return InternalPosition()


@dataclass
class SeparatorPosition(InternalPosition):
    pass


ArrowPoints   = List[InternalPosition]
DiamondPoints = List[InternalPosition]
PolygonPoints = Union[ArrowPoints, DiamondPoints]


@dataclass
class ScanPoints:
    """
    Used by diagramming methods that cannot fill in a polygon.  In that case, the diagrammer
    scans these points to determine if they are in the polygon.  If they are then presumably the
    diagramming method will draw a dot at the specified point to simulate a fill.
    """
    startScan: InternalPosition = field(default_factory=createInternalPositionFactory)
    endScan:   InternalPosition = field(default_factory=createInternalPositionFactory)




from typing import cast

from logging import Logger
from logging import getLogger

from xml.dom.minicompat import NodeList

from xml.dom.minidom import Document
from xml.dom.minidom import Element

from xml.dom.minidom import parseString

from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import ClassDefinitions
from pyumldiagrams.Definitions import DisplayMethodParameters
from pyumldiagrams.Definitions import LinePositions
from pyumldiagrams.Definitions import LineType
from pyumldiagrams.Definitions import MethodDefinition
from pyumldiagrams.Definitions import Methods
from pyumldiagrams.Definitions import ParameterDefinition
from pyumldiagrams.Definitions import Parameters
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import Size
from pyumldiagrams.Definitions import UmlLineDefinition
from pyumldiagrams.Definitions import UmlLineDefinitions
from pyumldiagrams.Definitions import createMethodsFactory

from pyumldiagrams.UnsupportedException import UnsupportedException

from pyumldiagrams.xmlsupport.XmlConstants import ATTR_DEFAULT_VALUE
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_DISPLAY_PARAMETERS
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_HEIGHT
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_LINK_DESTINATION_ANCHOR_X
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_LINK_DESTINATION_ANCHOR_Y
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_LINK_SOURCE_ANCHOR_X
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_LINK_SOURCE_ANCHOR_Y
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_NAME
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_SHOW_FIELDS
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_SHOW_METHODS
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_SHOW_STEREOTYPE
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_TYPE
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_WIDTH
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_X
from pyumldiagrams.xmlsupport.XmlConstants import ATTR_Y
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_CONTROL_POINT
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_GRAPHIC_CLASS
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_GRAPHIC_LINK
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_MODEL_CLASS
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_MODEL_LINK
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_MODEL_METHOD
from pyumldiagrams.xmlsupport.XmlConstants import ELEMENT_MODEL_PARAM


class ToClassDefinition:

    def __init__(self, fqFileName: str):

        self.logger: Logger = getLogger(__name__)

        self._xmlData:          str      = ''

        self._classDefinitions: ClassDefinitions   = []
        self._umlLineDefinitions:  UmlLineDefinitions = []

        self._documentNode: Document = cast(Document, None)
        with open(fqFileName) as xmlFile:
            self._xmlData = xmlFile.read()
            self._documentNode = parseString(self._xmlData)

    def generateClassDefinitions(self):

        graphicClassNodes: NodeList = self._documentNode.getElementsByTagName(ELEMENT_GRAPHIC_CLASS)

        self.logger.debug(f'{graphicClassNodes=}')

        for xmlGraphicClassNode in graphicClassNodes:

            xmlGraphicClass: Element = cast(Element, xmlGraphicClassNode)

            height: int = int(xmlGraphicClass.getAttribute(ATTR_HEIGHT))
            width:  int = int(xmlGraphicClass.getAttribute(ATTR_WIDTH))
            x:      int = int(xmlGraphicClass.getAttribute(ATTR_X))
            y:      int = int(xmlGraphicClass.getAttribute(ATTR_Y))

            xmlClass:  Element = xmlGraphicClass.getElementsByTagName(ELEMENT_MODEL_CLASS)[0]
            className: str     = xmlClass.getAttribute(ATTR_NAME)

            displayMethods:    bool = self._stringToBoolean(xmlClass.getAttribute(ATTR_SHOW_METHODS))
            displayFields:     bool = self._stringToBoolean(xmlClass.getAttribute(ATTR_SHOW_FIELDS))
            displayStereotype: bool = self._stringToBoolean(xmlClass.getAttribute(ATTR_SHOW_STEREOTYPE))

            displayParametersStr: str = xmlClass.getAttribute(ATTR_DISPLAY_PARAMETERS)

            if displayParametersStr is None or displayParametersStr == '':
                displayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.UNSPECIFIED
            else:
                displayMethodParameters = DisplayMethodParameters(displayParametersStr)

            classDef: ClassDefinition = ClassDefinition(name=className)

            classDef.displayMethods    = displayMethods
            classDef.displayFields     = displayFields
            classDef.displayStereotype = displayStereotype
            classDef.displayMethodParameters = displayMethodParameters

            classSize: Size = Size(width=width, height=height)
            classDef.size = classSize

            position: Position = Position(x=x, y=y)
            classDef.position = position

            classDef.methods = self.generateMethods(xmlClass=xmlClass)

            self.logger.debug(f'{classDef=}')
            self._classDefinitions.append(classDef)

    def generateMethods(self, xmlClass: Element) -> Methods:

        methods: Methods = createMethodsFactory()

        for xmlMethod in xmlClass.getElementsByTagName(ELEMENT_MODEL_METHOD):
            methodName: str = xmlMethod.getAttribute(ATTR_NAME)
            self.logger.debug(f'{methodName=}')

            method: MethodDefinition = MethodDefinition(name=methodName)

            method = self._generateMethodParameters(xmlMethod=xmlMethod, methodDef=method)

            methods.append(method)

        return methods

    def generateUmlLineDefinitions(self):

        graphicLinkNodes: NodeList = self._documentNode.getElementsByTagName(ELEMENT_GRAPHIC_LINK)

        for graphicLinkNode in graphicLinkNodes:

            xmlGraphicLink: Element = cast(Element, graphicLinkNode)

            xmlLink:       Element  = xmlGraphicLink.getElementsByTagName(ELEMENT_MODEL_LINK)[0]
            controlPoints: NodeList = xmlGraphicLink.getElementsByTagName(ELEMENT_CONTROL_POINT)

            srcX: int = int(xmlGraphicLink.getAttribute(ATTR_LINK_SOURCE_ANCHOR_X))
            srcY: int = int(xmlGraphicLink.getAttribute(ATTR_LINK_SOURCE_ANCHOR_Y))

            strType:  str      = xmlLink.getAttribute(ATTR_TYPE)
            lineType: LineType = LineType.toEnum(strType)

            srcPosition: Position = Position(x=srcX, y=srcY)
            linePositions: LinePositions = [srcPosition]
            umlLineDefinition: UmlLineDefinition = UmlLineDefinition(linePositions=linePositions, lineType=lineType)

            for point in controlPoints:

                controlPoint: Element = cast(Element, point)

                self.logger.debug(f'{controlPoint=}')
                x: int = int(controlPoint.getAttribute(ATTR_X))
                y: int = int(controlPoint.getAttribute(ATTR_Y))
                bendPosition: Position = Position(x=x, y=y)
                linePositions.append(bendPosition)

            destX: int = int(xmlGraphicLink.getAttribute(ATTR_LINK_DESTINATION_ANCHOR_X))
            destY: int = int(xmlGraphicLink.getAttribute(ATTR_LINK_DESTINATION_ANCHOR_Y))

            destPosition: Position = Position(x=destX, y=destY)

            linePositions.append(destPosition)
            self.logger.debug(f'{umlLineDefinition=}')
            self._umlLineDefinitions.append(umlLineDefinition)

    @property
    def classDefinitions(self) -> ClassDefinitions:
        return self._classDefinitions

    @classDefinitions.setter
    def classDefinitions(self, newDefinitions: ClassDefinitions):
        raise UnsupportedException('Class definitions are read-only')

    @property
    def umlLineDefinitions(self) -> UmlLineDefinitions:
        return self._umlLineDefinitions

    @umlLineDefinitions.setter
    def umlLineDefinitions(self, newDefinitions: UmlLineDefinitions):
        raise UnsupportedException('UML Line definitions are read-only')

    def _generateMethodParameters(self, xmlMethod: Element, methodDef: MethodDefinition) -> MethodDefinition:

        parameters: Parameters = Parameters([])
        for xmlParam in xmlMethod.getElementsByTagName(ELEMENT_MODEL_PARAM):
            paramDef: ParameterDefinition = self._getParam(xmlParam=xmlParam)
            parameters.append(paramDef)

        methodDef.parameters = parameters

        return methodDef

    def _getParam(self, xmlParam: Element) -> ParameterDefinition:

        paramName:    str = xmlParam.getAttribute(ATTR_NAME)
        paramType:    str = xmlParam.getAttribute(ATTR_TYPE)
        defaultValue: str = xmlParam.getAttribute(ATTR_DEFAULT_VALUE)
        self.logger.debug(f'{paramName=} {paramType=} {defaultValue=}')

        parameterDefinition: ParameterDefinition = ParameterDefinition(name=paramName,  parameterType=paramType)

        parameterDefinition.defaultValue = defaultValue

        return parameterDefinition

    def _stringToBoolean(self, strBoolValue: str) -> bool:

        self.logger.debug(f'{strBoolValue=}')
        try:
            if strBoolValue is not None:
                if strBoolValue in [True, "True", "true", 1, "1"]:
                    return True
        except (ValueError, Exception) as e:
            self.logger.error(f'_stringToBoolean error: {e}')

        return False




from typing import Final
from typing import Tuple
from typing import Union

from logging import Logger
from logging import getLogger

from os import sep as osSep

from codeallybasic.ResourceManager import ResourceManager

from PIL import Image
from PIL import ImageColor
from PIL import ImageDraw
from PIL import ImageFont


from pyumldiagrams.BaseDiagram import BaseDiagram
from pyumldiagrams.Definitions import DisplayMethodParameters

from pyumldiagrams.Definitions import TOP_MARGIN
from pyumldiagrams.Definitions import LEFT_MARGIN

from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import EllipseDefinition
from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import RectangleDefinition
from pyumldiagrams.Definitions import Size
from pyumldiagrams.Definitions import UmlLineDefinition

from pyumldiagrams.image.ImageCommon import ImageCommon

from pyumldiagrams.Internal import InternalPosition
from pyumldiagrams.Internal import SeparatorPosition

from pyumldiagrams.image.ImageFormat import ImageFormat
from pyumldiagrams.image.ImageLine import ImageLine

ShapeDefinition = Union[EllipseDefinition, RectangleDefinition]


class ImageDiagram(BaseDiagram):

    RESOURCES_PACKAGE_NAME: Final = 'pyumldiagrams.image.resources'
    RESOURCES_PATH:         Final = f'pyumldiagrams{osSep}image{osSep}resources'

    DEFAULT_IMAGE_WIDTH:  Final = 1280    # pixels
    DEFAULT_IMAGE_HEIGHT: Final = 1024    # pixels

    DEFAULT_BACKGROUND_COLOR: str = 'LightYellow'
    DEFAULT_LINE_COLOR:       str = 'Black'
    DEFAULT_TEXT_COLOR:       str = 'Black'
    DEFAULT_IMAGE_FORMAT:     str = ImageFormat.PNG.value
    SUFFIX_INDICATOR:         str = '.'

    X_NUDGE_FACTOR:        Final = 4
    Y_NUDGE_FACTOR:        Final = 6
    FIRST_METHOD_Y_OFFSET: Final = 0

    def __init__(self, fileName: str, docDisplayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.DISPLAY,
                 headerText: str = '',
                 imageSize: Size = Size(width=DEFAULT_IMAGE_WIDTH, height=DEFAULT_IMAGE_HEIGHT)):
        """
        Constructor

        Args:
            fileName:  The output file name.  Including the suffix

            docDisplayMethodParameters:  The global value to consult if a class value says UNSPECIFIED

            headerText:  The text to display as a header on the diagram

            imageSize:  The diagram size in pixels
        """

        super().__init__(fileName=fileName, docDisplayMethodParameters=docDisplayMethodParameters, headerText=headerText)

        self.logger: Logger = getLogger(__name__)

        self._img:   Image  = Image.new(mode='RGB',
                                        size=(imageSize.width, imageSize.height),
                                        color=ImageColor.getrgb(ImageDiagram.DEFAULT_BACKGROUND_COLOR))

        self._imgDraw:    ImageDraw = ImageDraw.Draw(self._img)
        self._lineDrawer: ImageLine = ImageLine(docWriter=self._imgDraw, diagramPadding=self._diagramPadding)

        # noinspection SpellCheckingInspection
        fqPath:     str       = self.retrieveResourcePath('MonoFonto.ttf')
        self._font:       ImageFont = ImageFont.truetype(font=fqPath, size=BaseDiagram.DEFAULT_FONT_SIZE)
        self._headerFont: ImageFont = ImageFont.truetype(font=fqPath, size=BaseDiagram.HEADER_FONT_SIZE)
        #
        # https://www.exiv2.org/tags.html
        #
        # TODO need to write the following EXIF TAGS
        #
        # 74  0112  Orientation    The image orientation viewed in terms of rows and columns.
        # 305 0131  Software       Name and version number of the software package(s) used to create the image
        # 316 013C  HostComputer   The computer and/or operating system in use at the time of image creation.
        # 270:      ImageDescription
        # 37510:    'UserComment'
        #

    def retrieveResourcePath(self, bareFileName: str) -> str:

        fqFileName: str = ResourceManager.retrieveResourcePath(bareFileName=bareFileName,
                                                               resourcePath=ImageDiagram.RESOURCES_PATH,
                                                               packageName=ImageDiagram.RESOURCES_PACKAGE_NAME)

        return fqFileName

    def drawClass(self, classDefinition: ClassDefinition):
        """
        Draw the class diagram defined by the input

        Overrides the empty base definition

        Args:
            classDefinition:    The class definition
        """
        self._drawClassSymbol(classDefinition=classDefinition)

        position: Position = classDefinition.position
        size:     Size     = classDefinition.size

        iPos: InternalPosition = ImageCommon.toInternal(position=position, horizontalGap=self.horizontalGap, verticalGap=self.verticalGap)

        self._drawClassName(classDefinition=classDefinition, rectX=iPos.x, rectY=iPos.y, symbolWidth=size.width)

        separatorPosition: SeparatorPosition = self._drawSeparator(rectX=iPos.x, rectY=iPos.y, shapeWidth=size.width)

        fieldReprs:  BaseDiagram.FieldsRepr  = self._buildFields(classDefinition.fields)

        fieldSeparatorPosition: SeparatorPosition = self._drawFields(fieldReprs=fieldReprs, separatorPosition=separatorPosition)
        methodSeparatorPosition = self._drawSeparator(rectX=iPos.x, rectY=fieldSeparatorPosition.y, shapeWidth=size.width)

        methodReprs: BaseDiagram.MethodsRepr = self._buildMethods(classDefinition.methods, classDefinition.displayMethodParameters)

        if classDefinition.displayMethods is True:
            self._drawMethods(methodReprs=methodReprs, separatorPosition=methodSeparatorPosition)

    def drawUmlLine(self, lineDefinition: UmlLineDefinition):
        """
        Draw the inheritance, aggregation, or composition lines that describe the relationships
        between the UML classes

        Overrides the empty base definition

        Args:
            lineDefinition:   A UML Line definition
        """
        self._lineDrawer.draw(lineDefinition=lineDefinition)

    def drawEllipse(self, definition: EllipseDefinition):
        """
        Draw a general purpose ellipse

        Overrides the empty base definition

        Args:
            definition:     It's definition
        """
        xy = self.__toInternalCoordinates(definition=definition)
        self._imgDraw.ellipse(xy=xy, fill=None, outline=ImageDiagram.DEFAULT_LINE_COLOR, width=1)

    def drawRectangle(self, definition: RectangleDefinition):
        """
        Draw a general purpose rectangle

        Overrides the empty base definition

        Args:
            definition:  The rectangle definition
        """

        xy = self.__toInternalCoordinates(definition=definition)
        self._imgDraw.rectangle(xy=xy, fill=None, outline=ImageDiagram.DEFAULT_LINE_COLOR, width=1)

    def write(self):
        """
        Call this method when you are done with placing the diagram onto the image document.

        Overrides the empty base definition
        """
        if self._headerText is not None and self._headerText != '':

            xy = (LEFT_MARGIN, TOP_MARGIN // 2)
            self._imgDraw.text(xy=xy, fill=ImageDiagram.DEFAULT_TEXT_COLOR, font=self._headerFont, text=self._headerText)

        adjustedFileName: str = self._addSuffix(fileName=self._fileName, suffix=ImageDiagram.DEFAULT_IMAGE_FORMAT)

        self.logger.info(f'{adjustedFileName=}')
        self._img.save(adjustedFileName, ImageDiagram.DEFAULT_IMAGE_FORMAT)

    def _drawClassSymbol(self, classDefinition: ClassDefinition):

        imgDraw: ImageDraw = self._imgDraw

        position: Position = classDefinition.position
        size:     Size     = classDefinition.size

        iPos: InternalPosition = ImageCommon.toInternal(position=position, horizontalGap=self.horizontalGap, verticalGap=self.verticalGap)

        x0 = iPos.x
        y0 = iPos.y
        x1 = x0 + size.width
        y1 = y0 + size.height
        xy = (x0, y0, x1, y1)
        self.logger.debug(f'Class Symbol {xy=}')
        imgDraw.rectangle(xy=xy, fill=None, outline=ImageDiagram.DEFAULT_LINE_COLOR, width=1)

    def _drawClassName(self, classDefinition: ClassDefinition, rectX: float, rectY: float, symbolWidth: float):

        imgDraw: ImageDraw = self._imgDraw

        left, top, right, bottom = imgDraw.textbbox(xy=(0, 0), text=classDefinition.name, font=self._font)
        nameWidth = right - left
        textX: float = rectX + ((symbolWidth / 2) - (nameWidth / 2))
        textY: float = rectY + (self._fontSize / 2)

        xy = (textX, textY)
        self.logger.debug(f'ClassName {xy=}')
        imgDraw.text(xy=xy, fill=ImageDiagram.DEFAULT_TEXT_COLOR, font=self._font, text=classDefinition.name)

    def _drawSeparator(self, rectX: int, rectY: int, shapeWidth: int) -> SeparatorPosition:
        """
        Draws the UML separators between the various part of the UML shape
        Does the computation to determine where it drew the separator

        Args:
            rectX: x position of symbol
            rectY: y position of symbol
            shapeWidth: The width of the symbol

        Returns:  Where it drew the separator
        """
        imgDraw: ImageDraw = self._imgDraw

        separatorX: int = rectX
        separatorY: int = rectY + self._fontSize + ImageDiagram.Y_NUDGE_FACTOR

        endX: int = rectX + shapeWidth

        xy = [separatorX, separatorY, endX, separatorY]
        self.logger.debug(f'Separator {xy=}')
        imgDraw.line(xy=xy, fill=ImageDiagram.DEFAULT_LINE_COLOR, width=1)

        return SeparatorPosition(separatorX, separatorY)

    def _drawFields(self, fieldReprs: BaseDiagram.FieldsRepr, separatorPosition: SeparatorPosition) -> SeparatorPosition:

        imgDraw: ImageDraw = self._imgDraw
        x: int = separatorPosition.x + ImageDiagram.X_NUDGE_FACTOR
        y: int = separatorPosition.y + ImageDiagram.Y_NUDGE_FACTOR
        for fieldRepr in fieldReprs:
            xy = (x, y)
            imgDraw.text(xy=xy, fill=ImageDiagram.DEFAULT_TEXT_COLOR, font=self._font, text=fieldRepr)
            y = y + self._fontSize + 2

        y = (y - self._fontSize) - 2   # Adjust for last addition
        return SeparatorPosition(x=x, y=y)

    def _drawMethods(self, methodReprs: BaseDiagram.MethodsRepr, separatorPosition: SeparatorPosition):

        imgDraw: ImageDraw = self._imgDraw

        x: int = separatorPosition.x + ImageDiagram.X_NUDGE_FACTOR
        y: int = separatorPosition.y + ImageDiagram.FIRST_METHOD_Y_OFFSET

        for methodRepr in methodReprs:

            xy = (x, y)
            imgDraw.text(xy=xy, fill=ImageDiagram.DEFAULT_TEXT_COLOR, font=self._font, text=methodRepr)
            y = y + self._fontSize

    def _addSuffix(self, fileName: str, suffix: str) -> str:

        result = fileName.find(f'{ImageDiagram.SUFFIX_INDICATOR}{suffix}')
        if result == -1:
            adjustedFileName: str = f'{fileName}{ImageDiagram.SUFFIX_INDICATOR}{suffix}'
        else:
            adjustedFileName = fileName
        return adjustedFileName

    def __toInternalCoordinates(self, definition: ShapeDefinition) -> Tuple[int, int, int, int]:

        pos:  Position = definition.position
        size: Size     = definition.size

        internalStart: InternalPosition = self.__toInternal(position=pos)
        internalEnd:   InternalPosition = self.__toInternal(position=Position(x=pos.x + size.width, y=pos.y + size.height))

        x1 = internalStart.x
        y1 = internalStart.y
        x2 = internalEnd.x
        y2 = internalEnd.y

        xy = (x1, y1, x2, y2)

        return xy

    def __toInternal(self, position: Position) -> InternalPosition:

        verticalGap:   int = self._diagramPadding.verticalGap
        horizontalGap: int = self._diagramPadding.horizontalGap

        iPos: InternalPosition = ImageCommon.toInternal(position, verticalGap=verticalGap, horizontalGap=horizontalGap)

        return iPos




from typing import Final
from typing import Tuple

from math import pi
from math import atan
from math import cos
from math import sin

from pyumldiagrams.Internal import ArrowPoints
from pyumldiagrams.Internal import DiamondPoints
from pyumldiagrams.Internal import InternalPosition


class Common:

    INHERITANCE_ARROW_HEIGHT: Final = 10
    DIAMOND_HEIGHT:           Final = 8

    @classmethod
    def computeTheArrowVertices(cls, src: InternalPosition, dest: InternalPosition)  -> ArrowPoints:
        """
        Draw an arrow at the end of the line source-dest.

        Args:
            src:  points of the segment
            dest:  points of the segment

        Returns:
            A list of positions that describes a diamond to draw
        """
        # x1: float = src.x
        # y1: float = src.y
        # x2: float = dest.x
        # y2: float = dest.y
        #
        # deltaX: float = x2 - x1
        # deltaY: float = y2 - y1
        deltaX, deltaY = Common.computeDeltaXDeltaY(src, dest)
        if abs(deltaX) < 0.01:   # vertical segment
            if deltaY > 0:
                alpha = -pi/2
            else:
                alpha = pi/2
        else:
            if deltaX == 0:
                alpha = pi/2
            else:
                alpha = atan(deltaY/deltaX)
        if deltaX > 0:
            alpha += pi

        pi_6: float = pi/6      # radians for 30 degree angle

        alpha1: float = alpha + pi_6
        alpha2: float = alpha - pi_6
        size:   float = Common.INHERITANCE_ARROW_HEIGHT
        x2: int = dest.x
        y2: int = dest.y
        #
        # The names for the left and right points are correct for upward facing arrows
        # They are inverted for downward facing arrows
        #
        arrowTip:   InternalPosition = InternalPosition(x2, y2)
        rightPoint: InternalPosition = InternalPosition(x2 + size * cos(alpha1), y2 + size * sin(alpha1))    # type: ignore
        leftPoint:  InternalPosition = InternalPosition(x2 + size * cos(alpha2), y2 + size * sin(alpha2))    # type: ignore

        points: ArrowPoints = [rightPoint, arrowTip, leftPoint]

        return points

    @classmethod
    def computeDiamondVertices(cls, src: InternalPosition, dest: InternalPosition) -> DiamondPoints:
        """
        Args:
            src:    The source point
            dest:   The destination point
        """
        pi_6: float = pi/6     # radians for 30 degree angle
        x2:   int   = dest.x
        y2:   int   = dest.y

        deltaX, deltaY = Common.computeDeltaXDeltaY(src, dest)

        if abs(deltaX) < 0.01:  # vertical segment
            if deltaY > 0:
                alpha = -pi/2
            else:
                alpha = pi/2
        else:
            if deltaX == 0:
                if deltaY > 0:
                    alpha = pi/2
                else:
                    alpha = 3 * pi / 2
            else:
                alpha = atan(deltaY/deltaX)
        if deltaX > 0:
            alpha += pi

        alpha1: float = alpha + pi_6
        alpha2: float = alpha - pi_6
        size:   int   = Common.DIAMOND_HEIGHT

        # noinspection PyListCreation
        points: DiamondPoints = []

        points.append((InternalPosition(x2 + size * cos(alpha1), y2 + size * sin(alpha1)))) # type: ignore
        points.append(InternalPosition(x2, y2))
        points.append(InternalPosition(x2 + size * cos(alpha2), y2 + size * sin(alpha2)))           # type: ignore
        points.append(InternalPosition(x2 + 2 * size * cos(alpha), y2 + 2 * size * sin(alpha)))     # type: ignore

        return points

    @classmethod
    def computeDeltaXDeltaY(cls, src: InternalPosition, dest: InternalPosition) -> Tuple[int, int]:

        x1: int = src.x
        y1: int = src.y
        x2: int = dest.x
        y2: int = dest.y

        deltaX: int = x2 - x1
        deltaY: int = y2 - y1

        return deltaX, deltaY

    @classmethod
    def computeMidPointOfBottomLine(cls, startPos: InternalPosition, endPos: InternalPosition) -> InternalPosition:
        """
        These two coordinates are the two end-points of the bottom leg of the inheritance arrow
        midPoint = (x1+x2/2, y1+y2/2)

        Args:
            startPos: start of line
            endPos:   end of line

        Returns:  Midpoint between startPos - endPos

        """
        x1: int = startPos.x
        y1: int = startPos.y
        x2: int = endPos.x
        y2: int = endPos.y

        midX: int = (x1 + x2) // 2
        midY: int = (y1 + y2) // 2

        return InternalPosition(midX, midY)




import pathlib

from setuptools import setup

from pyumldiagrams import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

README  = (HERE / "README.md").read_text()
LICENSE = (HERE / 'LICENSE').read_text()

setup(
    name='pyumldiagrams',
    version=__version__,
    author='Humberto A. Sanchez II',
    author_email='humberto.a.sanchez.ii@gmail.com',
    maintainer='Humberto A. Sanchez II',
    maintainer_email='humberto.a.sanchez.ii@gmail.com',
    description='Draw UML diagrams in various formats',
    long_description = README,
    long_description_content_type='text/markdown',
    license=LICENSE,
    url='https://github.com/hasii2011/pyumldiagrams',
    packages=[
        'pyumldiagrams',
        'pyumldiagrams.image', 'pyumldiagrams.image.resources',
        'pyumldiagrams.pdf',   'pyumldiagrams.pdf.resources',
        'pyumldiagrams.xmlsupport'
    ],
    package_data={
        'pyumldiagrams':                 ['py.typed'],
        'pyumldiagrams.image.resources': ['py.typed'],
        'pyumldiagrams.image':           ['py.typed'],
        'pyumldiagrams.pdf':             ['py.typed'],
        'pyumldiagrams.pdf.resources':   ['py.typed'],
        'pyumldiagrams.xmlsupport':      ['py.typed'],
    },
    install_requires=['fpdf2==2.7.5', 'Pillow==10.0.1', 'codeallybasic==0.5.2']
)




ELEMENT_DOCUMENT:      str = 'PyutDocument'
TOP_LEVEL_ELEMENT:     str = 'PyutProject'
ELEMENT_GRAPHIC_CLASS: str = 'GraphicClass'
ELEMENT_GRAPHIC_LINK:  str = 'GraphicLink'
ELEMENT_MODEL_CLASS:   str = 'Class'
ELEMENT_MODEL_LINK:    str = 'Link'
ELEMENT_CONTROL_POINT: str = 'ControlPoint'

ELEMENT_MODEL_METHOD: str = 'Method'

ATTR_WIDTH:  str = 'width'
ATTR_HEIGHT: str = 'height'
ATTR_X:      str = 'x'
ATTR_Y:      str = 'y'
ATTR_NAME:   str = 'name'

ATTR_LINK_SOURCE_ANCHOR_X: str = 'srcX'
ATTR_LINK_SOURCE_ANCHOR_Y: str = 'srcY'

ATTR_LINK_DESTINATION_ANCHOR_X: str = 'dstX'
ATTR_LINK_DESTINATION_ANCHOR_Y: str = 'dstY'

ATTR_TYPE: str = 'type'

ATTR_SHOW_METHODS:      str = 'showMethods'
ATTR_SHOW_FIELDS:       str = 'showFields'
ATTR_SHOW_STEREOTYPE:   str = 'showStereotype'

ATTR_DISPLAY_PARAMETERS: str = 'displayParameters'

ELEMENT_MODEL_PARAM: str = 'Param'
ATTR_DEFAULT_VALUE:  str = 'defaultValue'




from typing import Final

from logging import Logger
from logging import getLogger

from os import sep as osSep

from datetime import datetime

from codeallybasic.ResourceManager import ResourceManager

from pyumldiagrams.BaseDiagram import BaseDiagram
from pyumldiagrams.Defaults import DEFAULT_LINE_WIDTH
from pyumldiagrams.Definitions import DisplayMethodParameters
from pyumldiagrams.Internal import SeparatorPosition

from pyumldiagrams.Definitions import ClassDefinition
from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import EllipseDefinition
from pyumldiagrams.Definitions import UmlLineDefinition

from pyumldiagrams.Definitions import Position
from pyumldiagrams.Definitions import RectangleDefinition
from pyumldiagrams.Definitions import Size

from pyumldiagrams.pdf.PdfCommon import Coordinates
from pyumldiagrams.pdf.PdfCommon import Dimensions
from pyumldiagrams.pdf.PdfCommon import PdfCommon
from pyumldiagrams.pdf.PdfCommon import PdfShapeDefinition

from pyumldiagrams.pdf.PdfLine import PdfLine
from pyumldiagrams.pdf.FPDFExtended import FPDFExtended


class PdfDiagram(BaseDiagram):
    """

    Always lays out in portrait mode.  Currently, only supports UML classes with methods.  Only supports
    inheritance, composition, aggregation, and association lines.

    You are allowed to set the gap between UML classes both horizontally and vertically.  Also, you are allowed to
    specify the text font size
    """
    FPDF_DRAW: str = 'D'

    RESOURCES_PACKAGE_NAME: Final = 'pdf.resources'
    RESOURCES_PATH:         Final = f'pdf{osSep}resources'

    X_NUDGE_FACTOR: Final = 4
    Y_NUDGE_FACTOR: Final = 4

    FIRST_METHOD_Y_OFFSET: Final = 7

    def __init__(self, fileName: str, dpi: int, docDisplayMethodParameters: DisplayMethodParameters = DisplayMethodParameters.DISPLAY, headerText: str = ''):
        """

        Args:
            fileName:    Fully qualified file name
            dpi:         dots per inch for the display we are mapping from
            docDisplayMethodParameters:  The global value to consult if a class value says UNSPECIFIED
            headerText:  The header to place on the page
        """
        super().__init__(fileName=fileName, docDisplayMethodParameters=docDisplayMethodParameters, dpi=dpi, headerText=headerText)
        # self._fileName: str = fileName
        # self._dpi:      int = dpi
        self.logger: Logger = getLogger(__name__)

        pdf = FPDFExtended(headerText=headerText)
        pdf.add_page()

        pdf.set_display_mode(zoom='default', layout='single')

        pdf.set_line_width(DEFAULT_LINE_WIDTH)

        pdf.set_creator('Humberto A. Sanchez II - The Great')
        pdf.set_author('Humberto A. Sanchez II - The Great')

        pdf.set_font('Arial', size=BaseDiagram.DEFAULT_FONT_SIZE)
        pdf.headerText = headerText

        self._pdf:      FPDFExtended = pdf
        self._fontSize: int          = BaseDiagram.DEFAULT_FONT_SIZE

        diagramPadding:   DiagramPadding = DiagramPadding()
        self._lineDrawer: PdfLine = PdfLine(pdf=pdf, diagramPadding=diagramPadding, dpi=dpi)

        self._diagramPadding: DiagramPadding = diagramPadding

    @property
    def docTimeStamp(self) -> datetime:
        """
        Overrides the empty base implementation
        """
        return self._pdf.creation_date

    @docTimeStamp.setter
    def docTimeStamp(self, timeStamp: datetime):
        """
        Overrides the empty base implementation
        """
        self._pdf.creation_date = timeStamp

    def retrieveResourcePath(self, bareFileName: str) -> str:
        """
        Overrides the empty base implementation

        Args:
            bareFileName:

        Returns: a fully qualified name
        """
        fqFileName: str = ResourceManager.retrieveResourcePath(bareFileName=bareFileName,
                                                               resourcePath=PdfDiagram.RESOURCES_PATH,
                                                               packageName=PdfDiagram.RESOURCES_PACKAGE_NAME)

        return fqFileName

    def drawClass(self, classDefinition: ClassDefinition):
        """
        Draw the class diagram defined by the input

        Args:
            classDefinition:    The class definition
        """

        position:      Position = classDefinition.position
        verticalGap:   int      = self._diagramPadding.verticalGap
        horizontalGap: int      = self._diagramPadding.horizontalGap

        coordinates: Coordinates = PdfCommon.convertPosition(pos=position, dpi=self._dpi, verticalGap=verticalGap, horizontalGap=horizontalGap)
        x: int = coordinates.x
        y: int = coordinates.y
        self.logger.debug(f'x,y: ({x},{y})')

        methodReprs: BaseDiagram.MethodsRepr = self._buildMethods(classDefinition.methods, classDefinition.displayMethodParameters)
        fieldReprs:  BaseDiagram.FieldsRepr  = self._buildFields(classDefinition.fields)

        symbolWidth: int = self._drawClassSymbol(classDefinition, rectX=x, rectY=y)

        separatorPosition: SeparatorPosition = self._drawSeparator(rectX=x, rectY=y, shapeWidth=symbolWidth)
        fieldSeparatorPosition: SeparatorPosition = self._drawFields(fieldReprs=fieldReprs, separatorPosition=separatorPosition)

        methodSeparatorPosition: SeparatorPosition = self._drawSeparator(rectX=x, rectY=fieldSeparatorPosition.y, shapeWidth=symbolWidth)

        if classDefinition.displayMethods is True:
            self._drawMethods(methodReprs=methodReprs, separatorPosition=methodSeparatorPosition)

    def drawUmlLine(self, lineDefinition: UmlLineDefinition):
        """
        Draw the inheritance, aggregation, or composition lines that describe the relationships
        between the UML classes

        Args:
            lineDefinition:   A UML Line definition
        """
        self._lineDrawer.draw(lineDefinition=lineDefinition)

    def drawEllipse(self, definition: EllipseDefinition):
        """
        Draw a general purpose ellipse

        Args:
            definition:     It's definition
        """

        pdfShapeDefinition: PdfShapeDefinition = self.__convertDefinition(definition)

        x:      int = pdfShapeDefinition.coordinates.x
        y:      int = pdfShapeDefinition.coordinates.y
        width:  int = pdfShapeDefinition.dimensions.width
        height: int = pdfShapeDefinition.dimensions.height

        self._pdf.ellipse(x=x, y=y, w=width, h=height, style=definition.renderStyle.value)

    def drawRectangle(self, definition: RectangleDefinition):
        """
        Draw a general purpose rectangle

        Args:
            definition:  The rectangle definition

        """
        pdfShapeDefinition: PdfShapeDefinition = self.__convertDefinition(definition)

        x:      int = pdfShapeDefinition.coordinates.x
        y:      int = pdfShapeDefinition.coordinates.y
        width:  int = pdfShapeDefinition.dimensions.width
        height: int = pdfShapeDefinition.dimensions.height

        self._pdf.rect(x=x, y=y, w=width, h=height, style=definition.renderStyle.value)

    def drawText(self, position: Position, text: str):
        """
        Draw text at the input position.  The method will appropriately convert the
        position to PDF points

        Args:
            position:  The display's x, y position
            text:   The text to display
        """

        coordinates: Coordinates = PdfCommon.convertPosition(position, dpi=self._dpi, verticalGap=self.verticalGap, horizontalGap=self.horizontalGap)
        self._pdf.text(x=coordinates.x, y=coordinates.y, txt=text)

    def write(self):
        """
        Call this method when you are done with placing the diagram onto a PDF document.
        """
        self._pdf.output(self._fileName)

    def _drawClassSymbol(self, classDefinition: ClassDefinition, rectX: int, rectY: int) -> int:
        """
        Draws the UML Class symbol.

        Args:
            classDefinition:    The class definition
            rectX:      x position
            rectY:      y position

        Returns:  The computed UML symbol width
        """

        symbolWidth:  int = classDefinition.size.width
        symbolHeight: int = classDefinition.size.height

        size: Size = Size(width=symbolWidth, height=symbolHeight)

        dimensions: Dimensions = self.__convertSize(size=size)
        convertedWidth:  int = dimensions.width
        convertedHeight: int = dimensions.height
        #
        # The docs for rect are incorrect;  style is string
        # noinspection PyTypeChecker
        self._pdf.rect(x=rectX, y=rectY, w=convertedWidth, h=convertedHeight, style=PdfDiagram.FPDF_DRAW)

        nameWidth: int = self._pdf.get_string_width(classDefinition.name)
        textX: int = rectX + ((symbolWidth // 2) - (nameWidth // 2))
        textY: int = rectY + self._fontSize

        self._pdf.text(x=textX, y=textY, txt=classDefinition.name)

        return convertedWidth

    def _drawSeparator(self, rectX: int, rectY: int, shapeWidth: int) -> SeparatorPosition:
        """
        Draws the UML separator between the class name and the start of the class definition
        Does the computation to determine where it drew the separator

        Args:
            rectX: x position of symbol
            rectY: y position of symbol
            shapeWidth: The width of the symbol

        Returns:  Where it drew the separator

        """

        separatorX: int = rectX
        separatorY: int = rectY + self._fontSize + PdfDiagram.Y_NUDGE_FACTOR

        endX: int = rectX + shapeWidth

        self._pdf.line(x1=separatorX, y1=separatorY, x2=endX, y2=separatorY)

        return SeparatorPosition(separatorX, separatorY)

    def _drawMethods(self, methodReprs: BaseDiagram.MethodsRepr, separatorPosition: SeparatorPosition):

        x: int = separatorPosition.x + PdfDiagram.X_NUDGE_FACTOR
        y: int = separatorPosition.y + PdfDiagram.Y_NUDGE_FACTOR + PdfDiagram.FIRST_METHOD_Y_OFFSET

        for methodRepr in methodReprs:

            self._pdf.text(x=x, y=y, txt=methodRepr)

            y = y + self._fontSize

    def _drawFields(self, fieldReprs: BaseDiagram.FieldsRepr, separatorPosition: SeparatorPosition) -> SeparatorPosition:

        x: int = separatorPosition.x + PdfDiagram.X_NUDGE_FACTOR
        y: int = separatorPosition.y + PdfDiagram.Y_NUDGE_FACTOR + 8

        for fieldRepr in fieldReprs:
            self._pdf.text(x=x, y=y, txt=fieldRepr)
            y = y + self._fontSize + 2

        y = y - self._fontSize - 2  # adjust for last addition

        return SeparatorPosition(x=x, y=y)

    def __convertDefinition(self, definition: RectangleDefinition) -> PdfShapeDefinition:
        """

        Args:
            definition:

        Returns: A description of the shape
        """
        coordinates: Coordinates = PdfCommon.convertPosition(definition.position, dpi=self._dpi, verticalGap=self.verticalGap, horizontalGap=self.horizontalGap)
        dimensions:  Dimensions = self.__convertSize(definition.size)

        return PdfShapeDefinition(coordinates=coordinates, dimensions=dimensions)

    def __convertSize(self, size: Size) -> Dimensions:

        width:  int = PdfCommon.toPdfPoints(size.width, self._dpi)
        height: int = PdfCommon.toPdfPoints(size.height, self._dpi)

        return Dimensions(width=width, height=height)




from typing import Any

from abc import ABCMeta
from abc import abstractmethod

from logging import Logger
from logging import getLogger

from pyumldiagrams.Definitions import DiagramPadding
from pyumldiagrams.Definitions import UmlLineDefinition


class IDiagramLine(metaclass=ABCMeta):

    clsLogger: Logger = getLogger(__name__)

    def __init__(self, docMaker: Any, diagramPadding: DiagramPadding, dpi: int = 0):
        """

        Args:
            docMaker:  An instance of the document generation object.

            diagramPadding: Object that contains the observed margins and gaps.  See `pyumldiagrams.Definitions.DiagramPadding`
        """

        self._docMaker:       Any = docMaker
        self._dpi:            int = dpi
        self._diagramPadding: DiagramPadding  = diagramPadding

    @abstractmethod
    def draw(self, lineDefinition: UmlLineDefinition):
        """
        Draw the line described by the input parameter

        Must be overridden by implementors

        Args:
            lineDefinition:  Describes the line to draw
        """
        pass
